from pyquil.device._isa import (
    Edge,
    GateInfo,
    ISA,
    MeasureInfo,
    Qubit,
    THETA,
    gates_in_isa,
    isa_from_graph,
    isa_to_graph,
)
from pyquil.device._main import AbstractDevice, Device, NxDevice
from pyquil.device._specs import EdgeSpecs, QubitSpecs, Specs, specs_from_graph



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Module for facilitating connections to the QVM / QPU.
"""
import warnings
from typing import Any

__all__ = [
    "QVMConnection",
    "QVMCompiler",
    "QPUCompiler",
    "Device",
    "ForestConnection",
    "pyquil_protect",
    "WavefunctionSimulator",
    "QuantumComputer",
    "list_quantum_computers",
    "get_qc",
    "local_forest_runtime",
    "local_qvm",
    "QAM",
    "QVM",
    "QPU",
    "QPUConnection",
    "BenchmarkConnection",
    "get_benchmarker",
]

from pyquil.api._base_connection import ForestConnection
from pyquil.api._benchmark import BenchmarkConnection, get_benchmarker
from pyquil.api._compiler import QVMCompiler, QPUCompiler
from pyquil.api._error_reporting import pyquil_protect
from pyquil.api._qam import QAM
from pyquil.api._qpu import QPU
from pyquil.api._quantum_computer import (
    QuantumComputer,
    list_quantum_computers,
    get_qc,
    local_forest_runtime,
    local_qvm,
)
from pyquil.api._qvm import QVMConnection, QVM
from pyquil.api._wavefunction_simulator import WavefunctionSimulator
from pyquil.device._main import Device


class QPUConnection(QPU):
    def __init__(self, *args: Any, **kwargs: Any):
        warnings.warn(
            "QPUConnection's semantics have changed for Forest 2. Consider using "
            "pyquil.get_qc('...') instead of creating this object directly. "
            "Please consult the migration guide for full details.",
            DeprecationWarning,
        )
        super(QPU, self).__init__(*args, **kwargs)



##############################################################################
# Copyright 2016-2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
from warnings import warn
from typing import Callable, Mapping, Optional, Tuple, Union

import numpy as np

from pyquil.quilatom import (
    Addr,
    Expression,
    MemoryReference,
    MemoryReferenceDesignator,
    ParameterDesignator,
    QubitDesignator,
    unpack_classical_reg,
    unpack_qubit,
)
from pyquil.quilbase import (
    AbstractInstruction,
    Gate,
    Halt,
    Reset,
    ResetQubit,
    Measurement,
    Nop,
    Wait,
    ClassicalNeg,
    ClassicalNot,
    ClassicalAnd,
    ClassicalInclusiveOr,
    ClassicalExclusiveOr,
    ClassicalEqual,
    ClassicalGreaterEqual,
    ClassicalGreaterThan,
    ClassicalLessEqual,
    ClassicalLessThan,
    ClassicalMove,
    ClassicalExchange,
    ClassicalConvert,
    ClassicalLoad,
    ClassicalStore,
    ClassicalAdd,
    ClassicalSub,
    ClassicalMul,
    ClassicalDiv,
)


def unpack_reg_val_pair(
    classical_reg1: MemoryReferenceDesignator,
    classical_reg2: Union[MemoryReferenceDesignator, int, float],
) -> Tuple[MemoryReference, Union[MemoryReference, int, float]]:
    """
    Helper function for typechecking / type-coercing arguments to constructors for binary classical
    operators.

    :param classical_reg1: Specifier for the classical memory address to be modified.
    :param classical_reg2: Specifier for the second argument: a classical memory address or an
        immediate value.
    :return: A pair of pyQuil objects suitable for use as operands.
    """
    left = unpack_classical_reg(classical_reg1)
    if isinstance(classical_reg2, (float, int)):
        return left, classical_reg2
    return left, unpack_classical_reg(classical_reg2)


def prepare_ternary_operands(
    classical_reg1: MemoryReferenceDesignator,
    classical_reg2: MemoryReferenceDesignator,
    classical_reg3: Union[MemoryReferenceDesignator, int, float],
) -> Tuple[MemoryReference, MemoryReference, Union[MemoryReference, int, float]]:
    """
    Helper function for typechecking / type-coercing arguments to constructors for ternary
    classical operators.

    :param classical_reg1: Specifier for the classical memory address to be modified.
    :param classical_reg2: Specifier for the left operand: a classical memory address.
    :param classical_reg3: Specifier for the right operand: a classical memory address or an
        immediate value.
    :return: A triple of pyQuil objects suitable for use as operands.
    """
    if isinstance(classical_reg1, int):
        raise TypeError("Target operand of comparison must be a memory address")
    classical_reg1 = unpack_classical_reg(classical_reg1)
    if isinstance(classical_reg2, int):
        raise TypeError("Left operand of comparison must be a memory address")
    classical_reg2 = unpack_classical_reg(classical_reg2)
    if not isinstance(classical_reg3, (float, int)):
        classical_reg3 = unpack_classical_reg(classical_reg3)

    return classical_reg1, classical_reg2, classical_reg3


def I(qubit: QubitDesignator) -> Gate:
    """Produces the I identity gate::

        I = [1, 0]
            [0, 1]

    This gate is a single qubit identity gate.
    Note that this gate is different that the NOP instruction as noise channels
    are typically still applied during the duration of identity gates. Identities will
    also block parallelization like any other gate.

    :param qubit: The qubit apply the gate to.
    :returns: A Gate object.
    """
    return Gate(name="I", params=[], qubits=[unpack_qubit(qubit)])


def X(qubit: QubitDesignator) -> Gate:
    """Produces the X ("NOT") gate::

        X = [[0, 1],
             [1, 0]]

    This gate is a single qubit X-gate.

    :param qubit: The qubit apply the gate to.
    :returns: A Gate object.
    """
    return Gate(name="X", params=[], qubits=[unpack_qubit(qubit)])


def Y(qubit: QubitDesignator) -> Gate:
    """Produces the Y gate::

        Y = [[0, 0 - 1j],
             [0 + 1j, 0]]

    This gate is a single qubit Y-gate.

    :param qubit: The qubit apply the gate to.
    :returns: A Gate object.
    """
    return Gate(name="Y", params=[], qubits=[unpack_qubit(qubit)])


def Z(qubit: QubitDesignator) -> Gate:
    """Produces the Z gate::

        Z = [[1,  0],
             [0, -1]]

    This gate is a single qubit Z-gate.

    :param qubit: The qubit apply the gate to.
    :returns: A Gate object.
    """
    return Gate(name="Z", params=[], qubits=[unpack_qubit(qubit)])


def H(qubit: QubitDesignator) -> Gate:
    """Produces the Hadamard gate::

        H = (1 / sqrt(2)) * [[1,  1],
                             [1, -1]]

    Produces the H instruction. This gate is a single qubit Hadamard gate.

    :param qubit: The qubit apply the gate to.
    :returns: A Gate object.
    """
    return Gate(name="H", params=[], qubits=[unpack_qubit(qubit)])


def S(qubit: QubitDesignator) -> Gate:
    """Produces the S gate::

        S = [[1, 0],
             [0, 1j]]

    This gate is a single qubit S-gate.

    :param qubit: The qubit apply the gate to.
    :returns: A Gate object.
    """
    return Gate(name="S", params=[], qubits=[unpack_qubit(qubit)])


def T(qubit: QubitDesignator) -> Gate:
    """Produces the T gate::

        T = [[1, 0],
             [0, exp(1j * pi / 4)]]

    This gate is a single qubit T-gate. It is the same as RZ(pi/4).

    :param qubit: The qubit apply the gate to.
    :returns: A Gate object.
    """
    return Gate(name="T", params=[], qubits=[unpack_qubit(qubit)])


def RX(angle: ParameterDesignator, qubit: QubitDesignator) -> Gate:
    """Produces the RX gate::

        RX(phi) = [[cos(phi / 2), -1j * sin(phi / 2)],
                   [-1j * sin(phi / 2), cos(phi / 2)]]

    This gate is a single qubit X-rotation.

    :param angle: The angle to rotate around the x-axis on the bloch sphere.
    :param qubit: The qubit apply the gate to.
    :returns: A Gate object.
    """
    return Gate(name="RX", params=[angle], qubits=[unpack_qubit(qubit)])


def RY(angle: ParameterDesignator, qubit: QubitDesignator) -> Gate:
    """Produces the RY gate::

        RY(phi) = [[cos(phi / 2), -sin(phi / 2)],
                   [sin(phi / 2), cos(phi / 2)]]

    This gate is a single qubit Y-rotation.

    :param angle: The angle to rotate around the y-axis on the bloch sphere.
    :param qubit: The qubit apply the gate to.
    :returns: A Gate object.
    """
    return Gate(name="RY", params=[angle], qubits=[unpack_qubit(qubit)])


def RZ(angle: ParameterDesignator, qubit: QubitDesignator) -> Gate:
    """Produces the RZ gate::

        RZ(phi) = [[cos(phi / 2) - 1j * sin(phi / 2), 0]
                   [0, cos(phi / 2) + 1j * sin(phi / 2)]]

    This gate is a single qubit Z-rotation.

    :param angle: The angle to rotate around the z-axis on the bloch sphere.
    :param qubit: The qubit apply the gate to.
    :returns: A Gate object.
    """
    return Gate(name="RZ", params=[angle], qubits=[unpack_qubit(qubit)])


def PHASE(angle: ParameterDesignator, qubit: QubitDesignator) -> Gate:
    """Produces the PHASE gate::

        PHASE(phi) = [[1, 0],
                      [0, exp(1j * phi)]]

    This is the same as the RZ gate.

    :param angle: The angle to rotate around the z-axis on the bloch sphere.
    :param qubit: The qubit apply the gate to.
    :returns: A Gate object.
    """
    return Gate(name="PHASE", params=[angle], qubits=[unpack_qubit(qubit)])


def CZ(control: QubitDesignator, target: QubitDesignator) -> Gate:
    """Produces a controlled-Z gate::

        CZ = [[1, 0, 0,  0],
              [0, 1, 0,  0],
              [0, 0, 1,  0],
              [0, 0, 0, -1]]


    This gate applies to two qubit arguments to produce the controlled-Z gate instruction.

    :param control: The control qubit.
    :param target: The target qubit. The target qubit has an Z-gate applied to it if the control
        qubit is in the excited state.
    :returns: A Gate object.
    """
    return Gate(name="CZ", params=[], qubits=[unpack_qubit(q) for q in (control, target)])


def CNOT(control: QubitDesignator, target: QubitDesignator) -> Gate:
    """Produces a controlled-NOT (controlled-X) gate::

        CNOT = [[1, 0, 0, 0],
                [0, 1, 0, 0],
                [0, 0, 0, 1],
                [0, 0, 1, 0]]

    This gate applies to two qubit arguments to produce the controlled-not gate instruction.

    :param control: The control qubit.
    :param target: The target qubit. The target qubit has an X-gate applied to it if the control
        qubit is in the ``|1>`` state.
    :returns: A Gate object.
    """
    return Gate(name="CNOT", params=[], qubits=[unpack_qubit(q) for q in (control, target)])


def CCNOT(control1: QubitDesignator, control2: QubitDesignator, target: QubitDesignator) -> Gate:
    """Produces a doubly-controlled NOT gate::

        CCNOT = [[1, 0, 0, 0, 0, 0, 0, 0],
                 [0, 1, 0, 0, 0, 0, 0, 0],
                 [0, 0, 1, 0, 0, 0, 0, 0],
                 [0, 0, 0, 1, 0, 0, 0, 0],
                 [0, 0, 0, 0, 1, 0, 0, 0],
                 [0, 0, 0, 0, 0, 1, 0, 0],
                 [0, 0, 0, 0, 0, 0, 0, 1],
                 [0, 0, 0, 0, 0, 0, 1, 0]]

    This gate applies to three qubit arguments to produce the controlled-controlled-not gate
    instruction.

    :param control1: The first control qubit.
    :param control2: The second control qubit.
    :param target: The target qubit. The target qubit has an X-gate applied to it if both control
        qubits are in the excited state.
    :returns: A Gate object.
    """
    qubits = [unpack_qubit(q) for q in (control1, control2, target)]
    return Gate(name="CCNOT", params=[], qubits=qubits)


def CPHASE00(angle: ParameterDesignator, control: QubitDesignator, target: QubitDesignator) -> Gate:
    """Produces a controlled-phase gate that phases the ``|00>`` state::

        CPHASE00(phi) = diag([exp(1j * phi), 1, 1, 1])

    This gate applies to two qubit arguments to produce the variant of the controlled phase
    instruction that affects the state 00.

    :param angle: The input phase angle to apply when both qubits are in the ``|0>`` state.
    :param control: Qubit 1.
    :param target: Qubit 2.
    :returns: A Gate object.
    """
    qubits = [unpack_qubit(q) for q in (control, target)]
    return Gate(name="CPHASE00", params=[angle], qubits=qubits)


def CPHASE01(angle: ParameterDesignator, control: QubitDesignator, target: QubitDesignator) -> Gate:
    """Produces a controlled-phase gate that phases the ``|01>`` state::

        CPHASE01(phi) = diag([1.0, exp(1j * phi), 1.0, 1.0])

    This gate applies to two qubit arguments to produce the variant of the controlled phase
    instruction that affects the state 01.

    :param angle: The input phase angle to apply when q1 is in the ``|1>`` state and q2 is in
        the ``|0>`` state.
    :param control: Qubit 1.
    :param target: Qubit 2.
    :returns: A Gate object.
    """
    qubits = [unpack_qubit(q) for q in (control, target)]
    return Gate(name="CPHASE01", params=[angle], qubits=qubits)


def CPHASE10(angle: ParameterDesignator, control: QubitDesignator, target: QubitDesignator) -> Gate:
    """Produces a controlled-phase gate that phases the ``|10>`` state::

        CPHASE10(phi) = diag([1, 1, exp(1j * phi), 1])

    This gate applies to two qubit arguments to produce the variant of the controlled phase
    instruction that affects the state 10.

    :param angle: The input phase angle to apply when q2 is in the ``|1>`` state and q1 is in
        the ``|0>`` state.
    :param control: Qubit 1.
    :param target: Qubit 2.
    :returns: A Gate object.
    """
    qubits = [unpack_qubit(q) for q in (control, target)]
    return Gate(name="CPHASE10", params=[angle], qubits=qubits)


# NOTE: We don't use ParameterDesignator here because of the following Sphinx error. This error
# can be resolved by importing Expression, but then flake8 complains about an unused import:
#   Cannot resolve forward reference in type annotations of "pyquil.gates.CPHASE":
#   name 'Expression' is not defined
def CPHASE(
    angle: Union[Expression, MemoryReference, np.int_, int, float, complex],
    control: QubitDesignator,
    target: QubitDesignator,
) -> Gate:
    """Produces a controlled-phase instruction::

        CPHASE(phi) = diag([1, 1, 1, exp(1j * phi)])

    This gate applies to two qubit arguments to produce the variant of the controlled phase
    instruction that affects the state 11.

    Compare with the ``CPHASExx`` variants. This variant is the most common and does
    not have a suffix, although you can think of it as ``CPHASE11``.

    :param angle: The input phase angle to apply when both qubits are in the ``|1>`` state.
    :param control: Qubit 1.
    :param target: Qubit 2.
    :returns: A Gate object.
    """
    qubits = [unpack_qubit(q) for q in (control, target)]
    return Gate(name="CPHASE", params=[angle], qubits=qubits)


def SWAP(q1: QubitDesignator, q2: QubitDesignator) -> Gate:
    """Produces a SWAP gate which swaps the state of two qubits::

        SWAP = [[1, 0, 0, 0],
                [0, 0, 1, 0],
                [0, 1, 0, 0],
                [0, 0, 0, 1]]


    :param q1: Qubit 1.
    :param q2: Qubit 2.
    :returns: A Gate object.
    """
    return Gate(name="SWAP", params=[], qubits=[unpack_qubit(q) for q in (q1, q2)])


def CSWAP(control: QubitDesignator, target_1: QubitDesignator, target_2: QubitDesignator) -> Gate:
    """Produces a controlled-SWAP gate. This gate conditionally swaps the state of two qubits::

        CSWAP = [[1, 0, 0, 0, 0, 0, 0, 0],
                 [0, 1, 0, 0, 0, 0, 0, 0],
                 [0, 0, 1, 0, 0, 0, 0, 0],
                 [0, 0, 0, 1, 0, 0, 0, 0],
                 [0, 0, 0, 0, 1, 0, 0, 0],
                 [0, 0, 0, 0, 0, 0, 1, 0],
                 [0, 0, 0, 0, 0, 1, 0, 0],
                 [0, 0, 0, 0, 0, 0, 0, 1]]


    :param control: The control qubit.
    :param target_1: The first target qubit.
    :param target_2: The second target qubit. The two target states are swapped if the control is
        in the ``|1>`` state.
    """
    qubits = [unpack_qubit(q) for q in (control, target_1, target_2)]
    return Gate(name="CSWAP", params=[], qubits=qubits)


def ISWAP(q1: QubitDesignator, q2: QubitDesignator) -> Gate:
    """Produces an ISWAP gate::

        ISWAP = [[1, 0,  0,  0],
                 [0, 0,  1j, 0],
                 [0, 1j, 0,  0],
                 [0, 0,  0,  1]]

    This gate swaps the state of two qubits, applying a -i phase to q1 when it
    is in the 1 state and a -i phase to q2 when it is in the 0 state.

    :param q1: Qubit 1.
    :param q2: Qubit 2.
    :returns: A Gate object.
    """
    return Gate(name="ISWAP", params=[], qubits=[unpack_qubit(q) for q in (q1, q2)])


def PSWAP(angle: ParameterDesignator, q1: QubitDesignator, q2: QubitDesignator) -> Gate:
    """Produces a parameterized SWAP gate::

        PSWAP(phi) = [[1, 0,             0,             0],
                      [0, 0,             exp(1j * phi), 0],
                      [0, exp(1j * phi), 0,             0],
                      [0, 0,             0,             1]]


    :param angle: The angle of the phase to apply to the swapped states. This phase is applied to
        q1 when it is in the 1 state and to q2 when it is in the 0 state.
    :param q1: Qubit 1.
    :param q2: Qubit 2.
    :returns: A Gate object.
    """
    return Gate(name="PSWAP", params=[angle], qubits=[unpack_qubit(q) for q in (q1, q2)])


def XY(angle: ParameterDesignator, q1: QubitDesignator, q2: QubitDesignator) -> Gate:
    """Produces a parameterized ISWAP gate::

        XY(phi) = [[1,               0,               0, 0],
                   [0,      cos(phi/2), 1j * sin(phi/2), 0],
                   [0, 1j * sin(phi/2),      cos(phi/2), 0],
                   [0,               0,               0, 1]

    :param angle: The angle of the rotation to apply to the population 1 subspace.
    :param q1: Qubit 1.
    :param q2: Qubit 2.
    :returns: A Gate object.
    """
    return Gate(name="XY", params=[angle], qubits=[unpack_qubit(q) for q in (q1, q2)])


WAIT = Wait()
"""
This instruction tells the quantum computation to halt. Typically these is used while classical
memory is being manipulated by a CPU in a hybrid classical/quantum algorithm.

:returns: A Wait object.
"""


def RESET(qubit_index: Optional[QubitDesignator] = None) -> Union[Reset, ResetQubit]:
    """
    Reset all qubits or just one specific qubit.

    :param qubit_index: The qubit to reset.
        This can be a qubit's index, a Qubit, or a QubitPlaceholder.
        If None, reset all qubits.
    :returns: A Reset or ResetQubit Quil AST expression corresponding to a global or targeted
        reset, respectively.
    """
    if qubit_index is not None:
        return ResetQubit(unpack_qubit(qubit_index))
    else:
        return Reset()


NOP = Nop()
"""
This instruction applies no operation at that timestep. Typically these are ignored in error-models.

:returns: A Nop object.
"""

HALT = Halt()
"""
This instruction ends the program.

:returns: A Halt object.
"""


def MEASURE(
    qubit: QubitDesignator, classical_reg: Optional[MemoryReferenceDesignator]
) -> Measurement:
    """
    Produce a MEASURE instruction.

    :param qubit: The qubit to measure.
    :param classical_reg: The classical register to measure into, or None.
    :return: A Measurement instance.
    """
    qubit = unpack_qubit(qubit)
    if classical_reg is None:
        address = None
    elif isinstance(classical_reg, int):
        warn(
            "Indexing measurement addresses by integers is deprecated. "
            "Replacing this with the MemoryReference ro[i] instead."
        )
        address = MemoryReference("ro", classical_reg)
    else:
        address = unpack_classical_reg(classical_reg)
    return Measurement(qubit, address)


def TRUE(classical_reg: Union[MemoryReference, int]) -> ClassicalMove:
    """
    Produce a TRUE instruction.

    :param classical_reg: A classical register to modify.
    :return: An instruction object representing the equivalent MOVE.
    """
    warn("`TRUE a` has been deprecated. Use `MOVE a 1` instead.")
    if isinstance(classical_reg, int):
        classical_reg = Addr(classical_reg)
    return MOVE(classical_reg, 1)


def FALSE(classical_reg: Union[MemoryReference, int]) -> ClassicalMove:
    """
    Produce a FALSE instruction.

    :param classical_reg: A classical register to modify.
    :return: An instruction object representing the equivalent MOVE.
    """
    warn("`FALSE a` has been deprecated. Use `MOVE a 0` instead.")
    if isinstance(classical_reg, int):
        classical_reg = Addr(classical_reg)
    return MOVE(classical_reg, 0)


def NEG(classical_reg: MemoryReferenceDesignator) -> ClassicalNeg:
    """
    Produce a NEG instruction.

    :param classical_reg: A classical memory address to modify.
    :return: A ClassicalNeg instance.
    """
    return ClassicalNeg(unpack_classical_reg(classical_reg))


def NOT(classical_reg: MemoryReferenceDesignator) -> ClassicalNot:
    """
    Produce a NOT instruction.

    :param classical_reg: A classical register to modify.
    :return: A ClassicalNot instance.
    """
    return ClassicalNot(unpack_classical_reg(classical_reg))


def AND(
    classical_reg1: MemoryReferenceDesignator, classical_reg2: Union[MemoryReferenceDesignator, int]
) -> ClassicalAnd:
    """
    Produce an AND instruction.

    NOTE: The order of operands was reversed in pyQuil <=1.9 .

    :param classical_reg1: The first classical register, which gets modified.
    :param classical_reg2: The second classical register or immediate value.
    :return: A ClassicalAnd instance.
    """
    left, right = unpack_reg_val_pair(classical_reg1, classical_reg2)
    assert isinstance(right, (MemoryReference, int))  # placate mypy
    return ClassicalAnd(left, right)


def OR(
    classical_reg1: MemoryReferenceDesignator, classical_reg2: MemoryReferenceDesignator
) -> ClassicalInclusiveOr:
    """
    Produce an OR instruction.

    NOTE: Deprecated. Use IOR instead.

    :param classical_reg1: The first classical register.
    :param classical_reg2: The second classical register, which gets modified.
    :return: A ClassicalOr instance.
    """
    warn("OR has been deprecated. Use IOR, inclusive OR, instead.")
    return IOR(classical_reg2, classical_reg1)


def IOR(
    classical_reg1: MemoryReferenceDesignator, classical_reg2: Union[MemoryReferenceDesignator, int]
) -> ClassicalInclusiveOr:
    """
    Produce an inclusive OR instruction.

    :param classical_reg1: The first classical register, which gets modified.
    :param classical_reg2: The second classical register or immediate value.
    :return: A ClassicalInclusiveOr instance.
    """
    left, right = unpack_reg_val_pair(classical_reg1, classical_reg2)
    assert isinstance(right, (MemoryReference, int))  # placate mypy
    return ClassicalInclusiveOr(left, right)


def XOR(
    classical_reg1: MemoryReferenceDesignator, classical_reg2: Union[MemoryReferenceDesignator, int]
) -> ClassicalExclusiveOr:
    """
    Produce an exclusive OR instruction.

    :param classical_reg1: The first classical register, which gets modified.
    :param classical_reg2: The second classical register or immediate value.
    :return: A ClassicalExclusiveOr instance.
    """
    left, right = unpack_reg_val_pair(classical_reg1, classical_reg2)
    assert isinstance(right, (MemoryReference, int))  # placate mypy
    return ClassicalExclusiveOr(left, right)


def MOVE(
    classical_reg1: MemoryReferenceDesignator,
    classical_reg2: Union[MemoryReferenceDesignator, int, float],
) -> ClassicalMove:
    """
    Produce a MOVE instruction.

    :param classical_reg1: The first classical register, which gets modified.
    :param classical_reg2: The second classical register or immediate value.
    :return: A ClassicalMove instance.
    """
    left, right = unpack_reg_val_pair(classical_reg1, classical_reg2)
    return ClassicalMove(left, right)


def EXCHANGE(
    classical_reg1: MemoryReferenceDesignator, classical_reg2: MemoryReferenceDesignator
) -> ClassicalExchange:
    """
    Produce an EXCHANGE instruction.

    :param classical_reg1: The first classical register, which gets modified.
    :param classical_reg2: The second classical register, which gets modified.
    :return: A ClassicalExchange instance.
    """
    left = unpack_classical_reg(classical_reg1)
    right = unpack_classical_reg(classical_reg2)
    return ClassicalExchange(left, right)


def LOAD(
    target_reg: MemoryReferenceDesignator, region_name: str, offset_reg: MemoryReferenceDesignator
) -> ClassicalLoad:
    """
    Produce a LOAD instruction.

    :param target_reg: LOAD storage target.
    :param region_name: Named region of memory to load from.
    :param offset_reg: Offset into region of memory to load from. Must be a MemoryReference.
    :return: A ClassicalLoad instance.
    """
    return ClassicalLoad(
        unpack_classical_reg(target_reg), region_name, unpack_classical_reg(offset_reg)
    )


def STORE(
    region_name: str,
    offset_reg: MemoryReferenceDesignator,
    source: Union[MemoryReferenceDesignator, int, float],
) -> ClassicalStore:
    """
    Produce a STORE instruction.

    :param region_name: Named region of memory to store to.
    :param offset_reg: Offset into memory region. Must be a MemoryReference.
    :param source: Source data. Can be either a MemoryReference or a constant.
    :return: A ClassicalStore instance.
    """
    if not isinstance(source, int) and not isinstance(source, float):
        source = unpack_classical_reg(source)
    return ClassicalStore(region_name, unpack_classical_reg(offset_reg), source)


def CONVERT(
    classical_reg1: MemoryReferenceDesignator, classical_reg2: MemoryReferenceDesignator
) -> ClassicalConvert:
    """
    Produce a CONVERT instruction.

    :param classical_reg1: MemoryReference to store to.
    :param classical_reg2: MemoryReference to read from.
    :return: A ClassicalConvert instance.
    """
    return ClassicalConvert(
        unpack_classical_reg(classical_reg1), unpack_classical_reg(classical_reg2)
    )


def ADD(
    classical_reg: MemoryReferenceDesignator, right: Union[MemoryReferenceDesignator, int, float]
) -> ClassicalAdd:
    """
    Produce an ADD instruction.

    :param classical_reg: Left operand for the arithmetic operation. Also serves as the store
        target.
    :param right: Right operand for the arithmetic operation.
    :return: A ClassicalAdd instance.
    """
    left, right = unpack_reg_val_pair(classical_reg, right)
    return ClassicalAdd(left, right)


def SUB(
    classical_reg: MemoryReferenceDesignator, right: Union[MemoryReferenceDesignator, int, float]
) -> ClassicalSub:
    """
    Produce a SUB instruction.

    :param classical_reg: Left operand for the arithmetic operation. Also serves as the store
        target.
    :param right: Right operand for the arithmetic operation.
    :return: A ClassicalSub instance.
    """
    left, right = unpack_reg_val_pair(classical_reg, right)
    return ClassicalSub(left, right)


def MUL(
    classical_reg: MemoryReferenceDesignator, right: Union[MemoryReferenceDesignator, int, float]
) -> ClassicalMul:
    """
    Produce a MUL instruction.

    :param classical_reg: Left operand for the arithmetic operation. Also serves as the store
        target.
    :param right: Right operand for the arithmetic operation.
    :return: A ClassicalMul instance.
    """
    left, right = unpack_reg_val_pair(classical_reg, right)
    return ClassicalMul(left, right)


def DIV(
    classical_reg: MemoryReferenceDesignator, right: Union[MemoryReferenceDesignator, int, float]
) -> ClassicalDiv:
    """
    Produce an DIV instruction.

    :param classical_reg: Left operand for the arithmetic operation. Also serves as the store
        target.
    :param right: Right operand for the arithmetic operation.
    :return: A ClassicalDiv instance.
    """
    left, right = unpack_reg_val_pair(classical_reg, right)
    return ClassicalDiv(left, right)


def EQ(
    classical_reg1: MemoryReferenceDesignator,
    classical_reg2: MemoryReferenceDesignator,
    classical_reg3: Union[MemoryReferenceDesignator, int, float],
) -> ClassicalEqual:
    """
    Produce an EQ instruction.

    :param classical_reg1: Memory address to which to store the comparison result.
    :param classical_reg2: Left comparison operand.
    :param classical_reg3: Right comparison operand.
    :return: A ClassicalEqual instance.
    """
    classical_reg1, classical_reg2, classical_reg3 = prepare_ternary_operands(
        classical_reg1, classical_reg2, classical_reg3
    )

    return ClassicalEqual(classical_reg1, classical_reg2, classical_reg3)


def LT(
    classical_reg1: MemoryReferenceDesignator,
    classical_reg2: MemoryReferenceDesignator,
    classical_reg3: Union[MemoryReferenceDesignator, int, float],
) -> ClassicalLessThan:
    """
    Produce an LT instruction.

    :param classical_reg1: Memory address to which to store the comparison result.
    :param classical_reg2: Left comparison operand.
    :param classical_reg3: Right comparison operand.
    :return: A ClassicalLessThan instance.
    """
    classical_reg1, classical_reg2, classical_reg3 = prepare_ternary_operands(
        classical_reg1, classical_reg2, classical_reg3
    )
    return ClassicalLessThan(classical_reg1, classical_reg2, classical_reg3)


def LE(
    classical_reg1: MemoryReferenceDesignator,
    classical_reg2: MemoryReferenceDesignator,
    classical_reg3: Union[MemoryReferenceDesignator, int, float],
) -> ClassicalLessEqual:
    """
    Produce an LE instruction.

    :param classical_reg1: Memory address to which to store the comparison result.
    :param classical_reg2: Left comparison operand.
    :param classical_reg3: Right comparison operand.
    :return: A ClassicalLessEqual instance.
    """
    classical_reg1, classical_reg2, classical_reg3 = prepare_ternary_operands(
        classical_reg1, classical_reg2, classical_reg3
    )
    return ClassicalLessEqual(classical_reg1, classical_reg2, classical_reg3)


def GT(
    classical_reg1: MemoryReferenceDesignator,
    classical_reg2: MemoryReferenceDesignator,
    classical_reg3: Union[MemoryReferenceDesignator, int, float],
) -> ClassicalGreaterThan:
    """
    Produce an GT instruction.

    :param classical_reg1: Memory address to which to store the comparison result.
    :param classical_reg2: Left comparison operand.
    :param classical_reg3: Right comparison operand.
    :return: A ClassicalGreaterThan instance.
    """
    classical_reg1, classical_reg2, classical_reg3 = prepare_ternary_operands(
        classical_reg1, classical_reg2, classical_reg3
    )
    return ClassicalGreaterThan(classical_reg1, classical_reg2, classical_reg3)


def GE(
    classical_reg1: MemoryReferenceDesignator,
    classical_reg2: MemoryReferenceDesignator,
    classical_reg3: Union[MemoryReferenceDesignator, int, float],
) -> ClassicalGreaterEqual:
    """
    Produce an GE instruction.

    :param classical_reg1: Memory address to which to store the comparison result.
    :param classical_reg2: Left comparison operand.
    :param classical_reg3: Right comparison operand.
    :return: A ClassicalGreaterEqual instance.
    """
    classical_reg1, classical_reg2, classical_reg3 = prepare_ternary_operands(
        classical_reg1, classical_reg2, classical_reg3
    )
    return ClassicalGreaterEqual(classical_reg1, classical_reg2, classical_reg3)


QUANTUM_GATES: Mapping[str, Callable[..., Gate]] = {
    "I": I,
    "X": X,
    "Y": Y,
    "Z": Z,
    "H": H,
    "S": S,
    "T": T,
    "PHASE": PHASE,
    "RX": RX,
    "RY": RY,
    "RZ": RZ,
    "CZ": CZ,
    "CNOT": CNOT,
    "CCNOT": CCNOT,
    "CPHASE00": CPHASE00,
    "CPHASE01": CPHASE01,
    "CPHASE10": CPHASE10,
    "CPHASE": CPHASE,
    "SWAP": SWAP,
    "CSWAP": CSWAP,
    "ISWAP": ISWAP,
    "PSWAP": PSWAP,
    "XY": XY,
}
"""
Dictionary of quantum gate functions keyed by gate names.
"""

STANDARD_GATES = QUANTUM_GATES
"""
Alias for the above dictionary of quantum gates.
"""

STANDARD_INSTRUCTIONS: Mapping[
    str, Union[AbstractInstruction, Callable[..., AbstractInstruction]]
] = {
    "WAIT": WAIT,
    "RESET": RESET,
    "NOP": NOP,
    "HALT": HALT,
    "MEASURE": MEASURE,
    "TRUE": TRUE,
    "FALSE": FALSE,
    "NOT": NOT,
    "AND": AND,
    "OR": OR,
    "MOVE": MOVE,
    "EXCHANGE": EXCHANGE,
    "IOR": IOR,
    "XOR": XOR,
    "NEG": NEG,
    "ADD": ADD,
    "SUB": SUB,
    "MUL": MUL,
    "DIV": DIV,
    "EQ": EQ,
    "GT": GT,
    "GE": GE,
    "LE": LE,
    "LT": LT,
    "LOAD": LOAD,
    "STORE": STORE,
    "CONVERT": CONVERT,
}
"""
Dictionary of standard instruction functions keyed by instruction names.
"""

__all__ = (
    list(QUANTUM_GATES.keys())
    + list(STANDARD_INSTRUCTIONS.keys())
    + ["Gate", "QUANTUM_GATES", "STANDARD_GATES", "STANDARD_INSTRUCTIONS"]
)



##############################################################################
# Copyright 2016-2019 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
import sys
import warnings
from typing import Any, Dict, Optional, Sequence, Tuple

import networkx as nx

if sys.version_info < (3, 7):
    from pyquil.external.dataclasses import dataclass
else:
    from dataclasses import dataclass


@dataclass
class QubitSpecs:
    id: int
    fRO: Optional[float]
    f1QRB: Optional[float]
    f1QRB_std_err: Optional[float]
    f1Q_simultaneous_RB: Optional[float]
    f1Q_simultaneous_RB_std_err: Optional[float]
    T1: Optional[float]
    T2: Optional[float]
    fActiveReset: Optional[float]


@dataclass
class EdgeSpecs:
    targets: Tuple[int, ...]
    fBellState: Optional[float]
    fCZ: Optional[float]
    fCZ_std_err: Optional[float]
    fCPHASE: Optional[float]
    fCPHASE_std_err: Optional[float]
    fXY: Optional[float]
    fXY_std_err: Optional[float]
    fISWAP: Optional[float]
    fISWAP_std_err: Optional[float]


@dataclass
class Specs:
    """
    Basic specifications for the device, such as gate fidelities and coherence times.

    :ivar qubits_specs: The specs associated with individual qubits.
    :ivar edges_specs: The specs associated with edges, or qubit-qubit pairs.
    """

    qubits_specs: Sequence[QubitSpecs]
    edges_specs: Sequence[EdgeSpecs]

    def f1QRBs(self) -> Dict[int, Optional[float]]:
        """
        Get a dictionary of single-qubit randomized benchmarking fidelities (for individual gate
        operation, normalized to unity) from the specs, keyed by qubit index.

        :return: A dictionary of 1Q RB fidelities, normalized to unity.
        """
        return {qs.id: qs.f1QRB for qs in self.qubits_specs}

    def f1QRB_std_errs(self) -> Dict[int, Optional[float]]:
        """
        Get a dictionary of the standard errors of single-qubit randomized
        benchmarking fidelities (for individual gate operation, normalized to unity)
        from the specs, keyed by qubit index.

        :return: A dictionary of 1Q RB fidelity standard errors, normalized to unity.
        """
        return {qs.id: qs.f1QRB_std_err for qs in self.qubits_specs}

    def f1Q_simultaneous_RBs(self) -> Dict[int, Optional[float]]:
        """
        Get a dictionary of single-qubit randomized benchmarking fidelities (for simultaneous gate
        operation across the chip, normalized to unity) from the specs, keyed by qubit index.

        :return: A dictionary of simultaneous 1Q RB fidelities, normalized to unity.
        """
        return {qs.id: qs.f1Q_simultaneous_RB for qs in self.qubits_specs}

    def f1Q_simultaneous_RB_std_errs(self) -> Dict[int, Optional[float]]:
        """
        Get a dictionary of the standard errors of single-qubit randomized
        benchmarking fidelities (for simultaneous gate operation across the chip, normalized to
        unity) from the specs, keyed by qubit index.

        :return: A dictionary of simultaneous 1Q RB fidelity standard errors, normalized to unity.
        """
        return {qs.id: qs.f1Q_simultaneous_RB_std_err for qs in self.qubits_specs}

    def fROs(self) -> Dict[int, Optional[float]]:
        """
        Get a dictionary of single-qubit readout fidelities (normalized to unity)
        from the specs, keyed by qubit index.

        :return: A dictionary of RO fidelities, normalized to unity.
        """
        return {qs.id: qs.fRO for qs in self.qubits_specs}

    def fActiveResets(self) -> Dict[int, Optional[float]]:
        """
        Get a dictionary of single-qubit active reset fidelities (normalized to unity) from the
        specs, keyed by qubit index.

        :return: A dictionary of reset fidelities, normalized to unity.
        """
        return {qs.id: qs.fActiveReset for qs in self.qubits_specs}

    def T1s(self) -> Dict[int, Optional[float]]:
        """
        Get a dictionary of T1s (in seconds) from the specs, keyed by qubit index.

        :return: A dictionary of T1s, in seconds.
        """
        return {qs.id: qs.T1 for qs in self.qubits_specs}

    def T2s(self) -> Dict[int, Optional[float]]:
        """
        Get a dictionary of T2s (in seconds) from the specs, keyed by qubit index.

        :return: A dictionary of T2s, in seconds.
        """
        return {qs.id: qs.T2 for qs in self.qubits_specs}

    def fBellStates(self) -> Dict[Tuple[int, ...], Optional[float]]:
        """
        Get a dictionary of two-qubit Bell state fidelities (normalized to unity)
        from the specs, keyed by targets (qubit-qubit pairs).

        :return: A dictionary of Bell state fidelities, normalized to unity.
        """
        warnings.warn(
            DeprecationWarning(
                "fBellState device specs have been deprecated, and will "
                "be removed in release v2.13 (targeted for October 2019)"
            )
        )
        return {tuple(es.targets): es.fBellState for es in self.edges_specs}

    def fCZs(self) -> Dict[Tuple[int, ...], Optional[float]]:
        """
        Get a dictionary of CZ fidelities (normalized to unity) from the specs,
        keyed by targets (qubit-qubit pairs).

        :return: A dictionary of CZ fidelities, normalized to unity.
        """
        return {tuple(es.targets): es.fCZ for es in self.edges_specs}

    def fISWAPs(self) -> Dict[Tuple[int, ...], Optional[float]]:
        """
        Get a dictionary of ISWAP fidelities (normalized to unity) from the specs,
        keyed by targets (qubit-qubit pairs).

        :return: A dictionary of ISWAP fidelities, normalized to unity.
        """
        return {tuple(es.targets): es.fISWAP for es in self.edges_specs}

    def fISWAP_std_errs(self) -> Dict[Tuple[int, ...], Optional[float]]:
        """
        Get a dictionary of the standard errors of the ISWAP fidelities from the specs,
        keyed by targets (qubit-qubit pairs).

        :return: A dictionary of ISWAP fidelities, normalized to unity.
        """
        return {tuple(es.targets): es.fISWAP_std_err for es in self.edges_specs}

    def fXYs(self) -> Dict[Tuple[int, ...], Optional[float]]:
        """
        Get a dictionary of XY(pi) fidelities (normalized to unity) from the specs,
        keyed by targets (qubit-qubit pairs).

        :return: A dictionary of XY/2 fidelities, normalized to unity.
        """
        return {tuple(es.targets): es.fXY for es in self.edges_specs}

    def fXY_std_errs(self) -> Dict[Tuple[int, ...], Optional[float]]:
        """
        Get a dictionary of the standard errors of the XY fidelities from the specs,
        keyed by targets (qubit-qubit pairs).

        :return: A dictionary of XY fidelities, normalized to unity.
        """
        return {tuple(es.targets): es.fXY_std_err for es in self.edges_specs}

    def fCZ_std_errs(self) -> Dict[Tuple[int, ...], Optional[float]]:
        """
        Get a dictionary of the standard errors of the CZ fidelities from the specs,
        keyed by targets (qubit-qubit pairs).

        :return: A dictionary of CZ fidelities, normalized to unity.
        """
        return {tuple(es.targets): es.fCZ_std_err for es in self.edges_specs}

    def fCPHASEs(self) -> Dict[Tuple[int, ...], Optional[float]]:
        """
        Get a dictionary of CPHASE fidelities (normalized to unity) from the specs,
        keyed by targets (qubit-qubit pairs).

        :return: A dictionary of CPHASE fidelities, normalized to unity.
        """
        warnings.warn(
            DeprecationWarning(
                "fCPHASE device specs have been deprecated, and will "
                "be removed in release v2.13 (targeted for October 2019)"
            )
        )
        return {tuple(es.targets): es.fCPHASE for es in self.edges_specs}

    def to_dict(self) -> Dict[str, Any]:
        """
        Create a JSON-serializable representation of the device Specs.

        The dictionary representation is of the form::

            {
                '1Q': {
                    "0": {
                        "f1QRB": 0.99,
                        "f1QRB_std_err": 0.02,
                        "T1": 20e-6,
                        ...
                    },
                    "1": {
                        "f1QRB": 0.989,
                        "f1QRB_std_err": 0.015,
                        "T1": 19e-6,
                        ...
                    },
                    ...
                },
                '2Q': {
                    "1-4": {
                        "fBellState": 0.93,
                        "fCZ": 0.92,
                        "fCZ_std_err": 0.03,
                        "fCPHASE": 0.91
                    },
                    "1-5": {
                        "fBellState": 0.9,
                        "fCZ": 0.89,
                        "fCZ_std_err": 0.05,
                        "fCPHASE": 0.88
                    },
                    ...
                },
                ...
            }

        :return: A dctionary representation of self.
        """
        return {
            "1Q": {
                "{}".format(qs.id): {
                    "f1QRB": qs.f1QRB,
                    "f1QRB_std_err": qs.f1QRB_std_err,
                    "f1Q_simultaneous_RB": qs.f1Q_simultaneous_RB,
                    "f1Q_simultaneous_RB_std_err": qs.f1Q_simultaneous_RB_std_err,
                    "fRO": qs.fRO,
                    "T1": qs.T1,
                    "T2": qs.T2,
                    "fActiveReset": qs.fActiveReset,
                }
                for qs in self.qubits_specs
            },
            "2Q": {
                "{}-{}".format(*es.targets): {
                    "fBellState": es.fBellState,
                    "fCZ": es.fCZ,
                    "fCZ_std_err": es.fCZ_std_err,
                    "fCPHASE": es.fCPHASE,
                    "fCPHASE_std_err": es.fCPHASE_std_err,
                    "fXY": es.fXY,
                    "fXY_std_err": es.fXY_std_err,
                    "fISWAP": es.fISWAP,
                    "fISWAP_std_err": es.fISWAP_std_err,
                }
                for es in self.edges_specs
            },
        }

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "Specs":
        """
        Re-create the Specs from a dictionary representation.

        :param d: The dictionary representation.
        :return: The restored Specs.
        """
        return Specs(
            qubits_specs=sorted(
                [
                    QubitSpecs(
                        id=int(q),
                        fRO=qspecs.get("fRO"),
                        f1QRB=qspecs.get("f1QRB"),
                        f1QRB_std_err=qspecs.get("f1QRB_std_err"),
                        f1Q_simultaneous_RB=qspecs.get("f1Q_simultaneous_RB"),
                        f1Q_simultaneous_RB_std_err=qspecs.get("f1Q_simultaneous_RB_std_err"),
                        T1=qspecs.get("T1"),
                        T2=qspecs.get("T2"),
                        fActiveReset=qspecs.get("fActiveReset"),
                    )
                    for q, qspecs in d["1Q"].items()
                ],
                key=lambda qubit_specs: qubit_specs.id,
            ),
            edges_specs=sorted(
                [
                    EdgeSpecs(
                        targets=tuple(int(q) for q in e.split("-")),
                        fBellState=especs.get("fBellState"),
                        fCZ=especs.get("fCZ"),
                        fCZ_std_err=especs.get("fCZ_std_err"),
                        fCPHASE=especs.get("fCPHASE"),
                        fCPHASE_std_err=especs.get("fCPHASE_std_err"),
                        fXY=especs.get("fXY"),
                        fXY_std_err=especs.get("fXY_std_err"),
                        fISWAP=especs.get("fISWAP"),
                        fISWAP_std_err=especs.get("fISWAP_std_err"),
                    )
                    for e, especs in d["2Q"].items()
                ],
                key=lambda edge_specs: edge_specs.targets,
            ),
        )


def specs_from_graph(graph: nx.Graph) -> Specs:
    """
    Generate a Specs object from a NetworkX graph with placeholder values for the actual specs.

    :param graph: The graph
    """
    qspecs = [
        QubitSpecs(
            id=q,
            fRO=0.90,
            f1QRB=0.99,
            f1QRB_std_err=0.01,
            f1Q_simultaneous_RB=0.99,
            f1Q_simultaneous_RB_std_err=0.02,
            T1=30e-6,
            T2=30e-6,
            fActiveReset=0.99,
        )
        for q in graph.nodes
    ]
    especs = [
        EdgeSpecs(
            targets=(q1, q2),
            fBellState=0.90,
            fCZ=0.90,
            fCZ_std_err=0.05,
            fCPHASE=0.80,
            fCPHASE_std_err=0.05,
            fXY=0.86,
            fXY_std_err=0.05,
            fISWAP=0.90,
            fISWAP_std_err=0.05,
        )
        for q1, q2 in graph.edges
    ]
    return Specs(qspecs, especs)



##############################################################################
# Copyright 2016-2019 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Schema definition of an ExperimentResult, which encapsulates the outcome of a collection of
measurements that are aimed at estimating the expectation value of some observable.
"""
import logging
import sys
import warnings
from typing import Any, Dict, List, Optional, Union

import numpy as np

from pyquil.experiment._setting import ExperimentSetting

if sys.version_info < (3, 7):
    from pyquil.external.dataclasses import dataclass
else:
    from dataclasses import dataclass


log = logging.getLogger(__name__)


@dataclass(frozen=True)
class ExperimentResult:
    """An expectation and standard deviation for the measurement of one experiment setting
    in a tomographic experiment.

    In the case of readout error calibration, we also include
    expectation, standard deviation and count for the calibration results, as well as the
    expectation and standard deviation for the corrected results.
    """

    setting: ExperimentSetting
    expectation: Union[float, complex]
    total_counts: int
    std_err: Optional[Union[float, complex]] = None
    raw_expectation: Optional[Union[float, complex]] = None
    raw_std_err: Optional[float] = None
    calibration_expectation: Optional[Union[float, complex]] = None
    calibration_std_err: Optional[Union[float, complex]] = None
    calibration_counts: Optional[int] = None
    additional_results: Optional[List["ExperimentResult"]] = None

    def __init__(
        self,
        setting: ExperimentSetting,
        expectation: Union[float, complex],
        total_counts: int,
        stddev: Optional[Union[float, complex]] = None,
        std_err: Optional[Union[float, complex]] = None,
        raw_expectation: Optional[Union[float, complex]] = None,
        raw_stddev: Optional[Union[float, complex]] = None,
        raw_std_err: Optional[Union[float, complex]] = None,
        calibration_expectation: Optional[Union[float, complex]] = None,
        calibration_stddev: Optional[Union[float, complex]] = None,
        calibration_std_err: Optional[Union[float, complex]] = None,
        calibration_counts: Optional[int] = None,
        additional_results: Optional[List["ExperimentResult"]] = None,
    ):

        object.__setattr__(self, "setting", setting)
        object.__setattr__(self, "expectation", expectation)
        object.__setattr__(self, "total_counts", total_counts)
        object.__setattr__(self, "raw_expectation", raw_expectation)
        object.__setattr__(self, "calibration_expectation", calibration_expectation)
        object.__setattr__(self, "calibration_counts", calibration_counts)
        object.__setattr__(self, "additional_results", additional_results)

        if stddev is not None:
            warnings.warn("'stddev' has been renamed to 'std_err'")
            std_err = stddev
        object.__setattr__(self, "std_err", std_err)

        if raw_stddev is not None:
            warnings.warn("'raw_stddev' has been renamed to 'raw_std_err'")
            raw_std_err = raw_stddev
        object.__setattr__(self, "raw_std_err", raw_std_err)

        if calibration_stddev is not None:
            warnings.warn("'calibration_stddev' has been renamed to 'calibration_std_err'")
            calibration_std_err = calibration_stddev
        object.__setattr__(self, "calibration_std_err", calibration_std_err)

    def get_stddev(self) -> Optional[Union[float, complex]]:
        warnings.warn("'stddev' has been renamed to 'std_err'")
        return self.std_err

    def set_stddev(self, value: Union[float, complex]) -> None:
        warnings.warn("'stddev' has been renamed to 'std_err'")
        object.__setattr__(self, "std_err", value)

    stddev = property(get_stddev, set_stddev)

    def get_raw_stddev(self) -> Optional[float]:
        warnings.warn("'raw_stddev' has been renamed to 'raw_std_err'")
        return self.raw_std_err

    def set_raw_stddev(self, value: float) -> None:
        warnings.warn("'raw_stddev' has been renamed to 'raw_std_err'")
        object.__setattr__(self, "raw_std_err", value)

    raw_stddev = property(get_raw_stddev, set_raw_stddev)

    def get_calibration_stddev(self) -> Optional[Union[float, complex]]:
        warnings.warn("'calibration_stddev' has been renamed to 'calibration_std_err'")
        return self.calibration_std_err

    def set_calibration_stddev(self, value: Union[float, complex]) -> None:
        warnings.warn("'calibration_stddev' has been renamed to 'calibration_std_err'")
        object.__setattr__(self, "calibration_std_err", value)

    calibration_stddev = property(get_calibration_stddev, set_calibration_stddev)

    def __str__(self) -> str:
        return f"{self.setting}: {self.expectation} +- {self.std_err}"

    def __repr__(self) -> str:
        return f"ExperimentResult[{self}]"

    def serializable(self) -> Dict[str, Any]:
        return {
            "type": "ExperimentResult",
            "setting": self.setting,
            "expectation": self.expectation,
            "std_err": self.std_err,
            "total_counts": self.total_counts,
            "raw_expectation": self.raw_expectation,
            "raw_std_err": self.raw_std_err,
            "calibration_expectation": self.calibration_expectation,
            "calibration_std_err": self.calibration_std_err,
            "calibration_counts": self.calibration_counts,
        }


def bitstrings_to_expectations(
    bitstrings: np.ndarray, joint_expectations: Optional[List[List[int]]] = None
) -> np.ndarray:
    """
    Given an array of bitstrings (each of which is represented as an array of bits), map them to
    expectation values and return the desired joint expectation values. If no joint expectations
    are desired, then just the 1 -> -1, 0 -> 1 mapping is performed.

    :param bitstrings: Array of bitstrings to map.
    :param joint_expectations: Joint expectation values to calculate. Each entry is a list which
        contains the qubits to use in calculating the joint expectation value. Entries of length
        one just calculate single-qubit expectation values. Defaults to None, which is equivalent
        to the list of single-qubit expectations [[0], [1], ..., [n-1]] for bitstrings of length n.
    :return: An array of expectation values, of the same length as the array of bitstrings. The
        "width" could be different than the length of an individual bitstring (n) depending on
        the value of the ``joint_expectations`` parameter.
    """
    expectations: np.ndarray = 1 - 2 * bitstrings

    if joint_expectations is None:
        return expectations

    region_size = len(expectations[0])

    e = []
    for c in joint_expectations:
        where = np.zeros(region_size, dtype=bool)
        where[c] = True
        e.append(np.prod(expectations[:, where], axis=1))
    return np.stack(e, axis=-1)


def correct_experiment_result(
    result: ExperimentResult, calibration: ExperimentResult,
) -> ExperimentResult:
    """
    Given a raw, unmitigated result and its associated readout calibration, produce the result
    absent readout error.

    :param result: An ``ExperimentResult`` object with unmitigated readout error.
    :param calibration: An ``ExperimentResult`` object resulting from running readout calibration
        on the ``ExperimentSetting`` associated with the ``result`` parameter.
    :return: An ``ExperimentResult`` object corrected for symmetric readout error.
    """
    corrected_expectation = result.expectation / calibration.expectation

    # combine standard errors (are we assuming the counts are the same?)
    assert result.std_err is not None and calibration.std_err is not None
    corrected_variance = ratio_variance(
        result.expectation, result.std_err ** 2, calibration.expectation, calibration.std_err ** 2
    )

    # recursively apply to additional results
    additional_results = None
    if result.additional_results is not None and calibration.additional_results:
        assert len(result.additional_results) == len(calibration.additional_results)
        additional_results = [
            correct_experiment_result(r, c)
            for r, c in zip(result.additional_results, calibration.additional_results)
        ]

    return ExperimentResult(
        setting=result.setting,
        expectation=corrected_expectation,
        std_err=np.sqrt(corrected_variance).item(),
        total_counts=result.total_counts,
        raw_expectation=result.expectation,
        raw_std_err=result.std_err,
        calibration_expectation=calibration.expectation,
        calibration_std_err=calibration.std_err,
        calibration_counts=calibration.total_counts,
        additional_results=additional_results,
    )


def ratio_variance(
    a: Union[float, np.ndarray],
    var_a: Union[float, np.ndarray],
    b: Union[float, np.ndarray],
    var_b: Union[float, np.ndarray],
) -> Union[float, np.ndarray]:
    r"""
    Given random variables 'A' and 'B', compute the variance on the ratio Y = A/B. Denote the
    mean of the random variables as a = E[A] and b = E[B] while the variances are var_a = Var[A]
    and var_b = Var[B] and the covariance as Cov[A,B]. The following expression approximates the
    variance of Y

    Var[Y] \approx (a/b) ^2 * ( var_a /a^2 + var_b / b^2 - 2 * Cov[A,B]/(a*b) )

    We assume the covariance of A and B is negligible, resting on the assumption that A and B
    are independently measured. The expression above rests on the assumption that B is non-zero,
    an assumption which we expect to hold true in most cases, but makes no such assumptions
    about A. If we allow E[A] = 0, then calculating the expression above via numpy would complain
    about dividing by zero. Instead, we can re-write the above expression as

    Var[Y] \approx var_a /b^2 + (a^2 * var_b) / b^4

    where we have dropped the covariance term as noted above.

    See the following for more details:
      - https://doi.org/10.1002/(SICI)1097-0320(20000401)39:4<300::AID-CYTO8>3.0.CO;2-O
      - http://www.stat.cmu.edu/~hseltman/files/ratio.pdf
      - https://w.wiki/EMh

    :param a: Mean of 'A', to be used as the numerator in a ratio.
    :param var_a: Variance in 'A'
    :param b: Mean of 'B', to be used as the numerator in a ratio.
    :param var_b: Variance in 'B'
    """
    return var_a / b ** 2 + (a ** 2 * var_b) / b ** 4



#!/usr/bin/env python
##############################################################################
# Copyright 2016-2019 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

import sys

from setuptools import setup, find_packages

VERSION_ERROR = """
PyQuil 2.0+ requires Python 3

To install the most recent version with support for Python 2, make sure you
have pip >= 9.0 as well as setuptools >= 24.2:

    pip install pip setuptools --upgrade

Then you can either:

    - Install an older version of PyQuil via pip install pyquil<2.0
    - Upgrade your system to use Python 3.6
"""

if sys.version_info < (3, 6):
    raise ImportError(VERSION_ERROR)

with open("VERSION.txt", "r") as f:
    __version__ = f.read().strip()

# Specify UTF-8 to guard against systems that default to an ASCII locale.
with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

# save the source code in version.py
with open("pyquil/version.py", "r") as f:
    version_file_source = f.read()

# overwrite version.py in the source distribution
with open("pyquil/version.py", "w") as f:
    f.write(f"__version__ = '{__version__}'\n")

setup(
    name="pyquil",
    version=__version__,
    author="Rigetti Computing",
    author_email="softapps@rigetti.com",
    description="A Python library to generate Quantum Instruction Language (Quil) Programs.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/rigetti/pyquil.git",
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    license="Apache-2.0",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        # The minimum spec for a working pyquil install.
        # note to developers: this should be a subset of requirements.txt
        "numpy",
        "antlr4-python3-runtime>=4.7.2,<4.8",
        "requests",
        "networkx>=2.0.0",
        "rpcq>=3.0.0",
        # dependency of contextvars, which we vendor
        "immutables==0.6",
    ],
    extras_require={"latex": ["ipython"],
                    "tutorials": ["forest-benchmarking", "jupyter", "matplotlib", "seaborn",
                                  "pandas", "scipy", "tqdm"]},
    keywords="quantum quil programming hybrid",
    python_requires=">=3.6",
)

# restore version.py to its previous state
with open("pyquil/version.py", "w") as f:
    f.write(version_file_source)



##############################################################################
# Copyright 2016-2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
import re
import time
import warnings
from json.decoder import JSONDecodeError
from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple, Union, cast

import numpy as np
import requests
from requests.adapters import HTTPAdapter
from urllib3 import Retry

from pyquil.api._config import PyquilConfig
from pyquil.api._error_reporting import _record_call
from pyquil.api._errors import (
    error_mapping,
    ApiError,
    UserMessageError,
    UnknownApiError,
    TooManyQubitsError,
)
from pyquil.api._logger import logger
from pyquil.quil import Program
from pyquil.version import __version__
from pyquil.wavefunction import Wavefunction

TYPE_EXPECTATION = "expectation"
TYPE_MULTISHOT = "multishot"
TYPE_MULTISHOT_MEASURE = "multishot-measure"
TYPE_WAVEFUNCTION = "wavefunction"


def get_json(session: requests.Session, url: str, params: Optional[Dict[Any, Any]] = None) -> Any:
    """
    Get JSON from a Forest endpoint.
    """
    logger.debug("Sending GET request to %s. Params: %s", url, params)
    res = session.get(url, params=params)
    if res.status_code >= 400:
        raise parse_error(res)
    return res.json()


def post_json(session: requests.Session, url: str, json: Any) -> requests.models.Response:
    """
    Post JSON to the Forest endpoint.
    """
    logger.debug("Sending POST request to %s. Body: %s", url, json)
    res = session.post(url, json=json)
    if res.status_code >= 400:
        raise parse_error(res)
    return res


def parse_error(res: requests.Response) -> ApiError:
    """
    Every server error should contain a "status" field with a human readable explanation of
    what went wrong as well as a "error_type" field indicating the kind of error that can be mapped
    to a Python type.

    There's a fallback error UnknownError for other types of exceptions (network issues, api
    gateway problems, etc.)
    """
    try:
        body = res.json()
    except JSONDecodeError:
        raise UnknownApiError(res.text)

    if "error_type" not in body:
        raise UnknownApiError(str(body))

    error_type = body["error_type"]
    status = body["status"]

    if re.search(r"[0-9]+ qubits were requested, but the QVM is limited to [0-9]+ qubits.", status):
        return TooManyQubitsError(status)

    error_cls = error_mapping.get(error_type, UnknownApiError)
    return error_cls(status)


def get_session(*args: Any, **kwargs: Any) -> "ForestSession":
    """
    Create a requests session to access the REST API

    :return: requests session
    :rtype: Session
    """
    session = ForestSession(*args, **kwargs)
    retry_adapter = HTTPAdapter(
        max_retries=Retry(
            total=3,
            method_whitelist=["POST"],
            status_forcelist=[502, 503, 504, 521, 523],
            backoff_factor=0.2,
            raise_on_status=False,
        )
    )

    session.mount("http://", retry_adapter)
    session.mount("https://", retry_adapter)

    # We need this to get binary payload for the wavefunction call.
    session.headers.update({"Accept": "application/octet-stream"})

    session.headers.update({"Content-Type": "application/json; charset=utf-8"})

    return session


def validate_noise_probabilities(noise_parameter: Optional[List[float]]) -> None:
    """
    Is noise_parameter a valid specification of noise probabilities for depolarizing noise?

    :param list noise_parameter: List of noise parameter values to be validated.
    """
    if not noise_parameter:
        return
    if not isinstance(noise_parameter, list):
        raise TypeError("noise_parameter must be a list")
    if any([not isinstance(value, float) for value in noise_parameter]):
        raise TypeError("noise_parameter values should all be floats")
    if len(noise_parameter) != 3:
        raise ValueError("noise_parameter lists must be of length 3")
    if sum(noise_parameter) > 1 or sum(noise_parameter) < 0:
        raise ValueError("sum of entries in noise_parameter must be between 0 and 1 (inclusive)")
    if any([value < 0 for value in noise_parameter]):
        raise ValueError("noise_parameter values should all be non-negative")


def validate_qubit_list(qubit_list: Sequence[int]) -> Sequence[int]:
    """
    Check the validity of qubits for the payload.

    :param qubit_list: List of qubits to be validated.
    """
    if not isinstance(qubit_list, Sequence):
        raise TypeError("'qubit_list' must be of type 'Sequence'")
    if any(not isinstance(i, int) or i < 0 for i in qubit_list):
        raise TypeError("'qubit_list' must contain positive integer values")
    return qubit_list


def prepare_register_list(
    register_dict: Dict[str, Union[bool, Sequence[int]]]
) -> Dict[str, Union[bool, Sequence[int]]]:
    """
    Canonicalize classical addresses for the payload and ready MemoryReference instances
    for serialization.

    This function will cast keys that are iterables of int-likes to a list of Python
    ints. This is to support specifying the register offsets as ``range()`` or numpy
    arrays. This mutates ``register_dict``.

    :param register_dict: The classical memory to retrieve. Specified as a dictionary:
        the keys are the names of memory regions, and the values are either (1) a list of
        integers for reading out specific entries in that memory region, or (2) True, for
        reading out the entire memory region.
    """
    if not isinstance(register_dict, dict):
        raise TypeError("register_dict must be a dict but got " + repr(register_dict))

    for k, v in register_dict.items():
        if isinstance(v, bool):
            assert v  # If boolean v must be True
            continue

        indices = [int(x) for x in v]  # support ranges, numpy, ...

        if not all(x >= 0 for x in indices):
            raise TypeError("Negative indices into classical arrays are not allowed.")
        register_dict[k] = indices

    return register_dict


def run_and_measure_payload(
    quil_program: Program, qubits: Sequence[int], trials: int, random_seed: int
) -> Dict[str, object]:
    """REST payload for :py:func:`ForestConnection._run_and_measure`"""
    if not quil_program:
        raise ValueError(
            "You have attempted to run an empty program."
            " Please provide gates or measure instructions to your program."
        )

    if not isinstance(quil_program, Program):
        raise TypeError("quil_program must be a Quil program object")
    qubits = validate_qubit_list(qubits)
    if not isinstance(trials, int):
        raise TypeError("trials must be an integer")

    payload = {
        "type": TYPE_MULTISHOT_MEASURE,
        "qubits": list(qubits),
        "trials": trials,
        "compiled-quil": quil_program.out(),
    }

    if random_seed is not None:
        payload["rng-seed"] = random_seed

    return payload


def wavefunction_payload(quil_program: Program, random_seed: int) -> Dict[str, object]:
    """REST payload for :py:func:`ForestConnection._wavefunction`"""
    if not isinstance(quil_program, Program):
        raise TypeError("quil_program must be a Quil program object")

    payload: Dict[str, object] = {"type": TYPE_WAVEFUNCTION, "compiled-quil": quil_program.out()}

    if random_seed is not None:
        payload["rng-seed"] = random_seed

    return payload


def expectation_payload(
    prep_prog: Program, operator_programs: Optional[Iterable[Program]], random_seed: int
) -> Dict[str, object]:
    """REST payload for :py:func:`ForestConnection._expectation`"""
    if operator_programs is None:
        operator_programs = [Program()]

    if not isinstance(prep_prog, Program):
        raise TypeError("prep_prog variable must be a Quil program object")

    payload: Dict[str, object] = {
        "type": TYPE_EXPECTATION,
        "state-preparation": prep_prog.out(),
        "operators": [x.out() for x in operator_programs],
    }

    if random_seed is not None:
        payload["rng-seed"] = random_seed

    return payload


def qvm_run_payload(
    quil_program: Program,
    classical_addresses: Dict[str, Union[bool, Sequence[int]]],
    trials: int,
    measurement_noise: Optional[Tuple[float, float, float]],
    gate_noise: Optional[Tuple[float, float, float]],
    random_seed: Optional[int],
) -> Dict[str, object]:
    """REST payload for :py:func:`ForestConnection._qvm_run`"""
    if not quil_program:
        raise ValueError(
            "You have attempted to run an empty program."
            " Please provide gates or measure instructions to your program."
        )
    if not isinstance(quil_program, Program):
        raise TypeError("quil_program must be a Quil program object")
    classical_addresses = prepare_register_list(classical_addresses)
    if not isinstance(trials, int):
        raise TypeError("trials must be an integer")

    payload = {
        "type": TYPE_MULTISHOT,
        "addresses": classical_addresses,
        "trials": trials,
        "compiled-quil": quil_program.out(),
    }

    if measurement_noise is not None:
        payload["measurement-noise"] = measurement_noise
    if gate_noise is not None:
        payload["gate-noise"] = gate_noise
    if random_seed is not None:
        payload["rng-seed"] = random_seed

    return payload


class ForestSession(requests.Session):
    """
    ForestSession inherits from requests.Session. It is responsible for adding
    authentication headers to Forest server requests. Upon receiving a 401 or 403
    response, it will attempt to refresh the auth credential and update the
    PyquilConfig, which in turn writes the refreshed auth credential to file.

    Encapsulates the operations required for authorization & encryption
    with the QPU.

    Two operations are involved in authorization:

    * Requesting & storing a user authentication token, used to authenticate calls
      to Forest, Dispatch, and other Rigetti services
    * Requesting a Curve ZeroMQ keypair for connection to the QPU. The response to
      this request also comes with service endpoints: compiler server and QPU

    The authentication tokens are of the standard JWT format and are issued by Forest Server.

    The refresh token is only used to renew the access token, which is used for all transactions
    and is valid for a short period of time.

    In wrapping the PyQuilConfig object, it provides that object with a callback to
    retrieve a valid engagement when needed, because the engagement is maintained here
    but is used by the config to provide service endpoints.
    """

    def __init__(self, *, config: PyquilConfig, lattice_name: Optional[str] = None):
        super().__init__()
        self.config = config
        self.config.get_engagement = self.get_engagement
        self._engagement: Optional["Engagement"] = None
        self.headers.update(self.config.qcs_auth_headers)
        self.headers["User-Agent"] = f"PyQuil/{__version__}"
        self.lattice_name = lattice_name

    def _engage(self) -> Optional["Engagement"]:
        """
        The heart of the QPU authorization process, ``engage`` makes a request to
        the dispatch server for the information needed to communicate with the QPU.

        This is a standard GraphQL request, authenticated using the access token
        retrieved from Forest Server.

        The response includes the endpoints to the QPU and QPU Compiler Server,
        along with the set of keys necessary to connect to the QPU and the time at
        which that key set expires.
        """
        query = """
          mutation Engage($name: String!) {
            engage(input: { lattice: { name: $name }}) {
              success
              message
              engagement {
                type
                qpu {
                    endpoint
                    credentials {
                        clientPublic
                        clientSecret
                        serverPublic
                    }
                }
                compiler {
                    endpoint
                }
                expiresAt
              }
            }
          }
        """
        if not self.lattice_name:
            logger.debug("ForestSession requires lattice_name in order to engage")
            return None

        logger.debug("Requesting engagement from %s", self.config.dispatch_url)
        variables = dict(name=self.lattice_name)
        query_response = self._request_graphql_retry(
            self.config.dispatch_url, query=query, variables=variables
        )

        if query_response.get("errors"):
            errors = query_response.get("errors", [])
            error_messages = map(lambda error: error["message"], errors)  # type: ignore
            raise UserMessageError(f"Failed to engage: {','.join(error_messages)}")

        engagement_response = query_response.get("data", {}).get("engage", None)
        if engagement_response and engagement_response.get("success") is True:
            logger.debug("Engagement successful")
            engagement_data = engagement_response.get("engagement", {})
            return Engagement(
                client_secret_key=engagement_data.get("qpu", {})
                .get("credentials", {})
                .get("clientSecret", "")
                .encode("utf-8"),
                client_public_key=engagement_data.get("qpu", {})
                .get("credentials", {})
                .get("clientPublic", "")
                .encode("utf-8"),
                server_public_key=engagement_data.get("qpu", {})
                .get("credentials", {})
                .get("serverPublic", "")
                .encode("utf-8"),
                expires_at=engagement_data.get("expiresAt", {}),
                qpu_endpoint=engagement_data.get("qpu", {}).get("endpoint"),
                qpu_compiler_endpoint=engagement_data.get("compiler", {}).get("endpoint"),
            )
        else:
            raise UserMessageError(
                f"Unable to engage {self.lattice_name}: "
                f"{engagement_response.get('message', 'No message')}"
            )

    def get_engagement(self) -> Optional["Engagement"]:
        """
        Returns memoized engagement information, if still valid - or requests a new engagement
        and then stores and returns that.
        """
        if not (self._engagement and self._engagement.is_valid()):
            self._engagement = self._engage()
        return self._engagement

    def _refresh_auth_token(self) -> bool:
        self.config.assert_valid_auth_credential()
        if self.config.user_auth_token is not None:
            return self._refresh_user_auth_token()
        elif self.config.qmi_auth_token is not None:
            return self._refresh_qmi_auth_token()
        return False

    def _refresh_user_auth_token(self) -> bool:
        url = f"{self.config.forest_url}/auth/idp/oauth2/v1/token"
        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "Cache-Control": "no-cache",
            "Accept": "application/json",
        }
        assert self.config.user_auth_token is not None
        data = {
            "grant_type": "refresh_token",
            "scope": self.config.user_auth_token["scope"],
            "refresh_token": self.config.user_auth_token["refresh_token"],
        }
        response = super().request("POST", url, data=data, headers=headers)
        if response.status_code == 200:
            self.config.update_user_auth_token(response.json())
            self.headers.update(self.config.qcs_auth_headers)
            return True

        logger.warning(
            f"Failed to refresh your user auth token at {self.config.user_auth_token_path}. "
            f"Server response: {response.text}"
        )
        return False

    def _refresh_qmi_auth_token(self) -> bool:
        url = f"{self.config.forest_url}/auth/qmi/refresh"
        headers = {"Content-Type": "application/json", "Accept": "application/json"}
        data = self.config.qmi_auth_token
        response = super().request("POST", url, json=data, headers=headers)
        if response.status_code == 200:
            self.config.update_qmi_auth_token(response.json())
            self.headers.update(self.config.qcs_auth_headers)
            return True

        logger.warning(
            f"Failed to refresh your QMI auth token at {self.config.qmi_auth_token_path}. "
            f"Server response: {response.text}"
        )
        return False

    def request(self, *args: Any, **kwargs: Any) -> requests.models.Response:
        """
        request is a wrapper around requests.Session#request that checks for
        401 and 403 response statuses and refreshes the auth credential
        accordingly.
        """
        response = super().request(*args, **kwargs)
        if response.status_code in {401, 403}:
            if self._refresh_auth_token():
                response = super().request(*args, **kwargs)
        return response

    def _request_graphql(self, url: str, query: str, variables: Dict[Any, Any]) -> Dict[Any, Any]:
        """
        Makes a single graphql request using the session credentials, throwing an error
        if the response is not valid JSON.

        Returns the JSON parsed from the response.
        """
        response = super().post(url, json=dict(query=query, variables=variables))
        try:
            return cast(Dict[Any, Any], response.json())
        except JSONDecodeError as e:
            logger.exception(f"Unable to parse json response from endpoint {url}:", response.text)
            raise e

    def _request_graphql_retry(self, *args: Any, **kwargs: Any) -> Dict[Any, Any]:
        """
        Makes a GraphQL request using session credentials, refreshing them once if the server
        identifies them as expired.

        Determining whether a call has failed to a GraphQL endpoint is less axiomatic than for a
        REST interface, and so here we follow the pattern set by Rigetti services, which return an
        HTTP 200 response with an array of errors. If any of those errors cite an expired
        authentication token, we refresh the token to clear that error. Note that other error
        messages will not trigger a retry.
        """
        result = self._request_graphql(*args, **kwargs)
        errors = result.get("errors", [])
        token_is_expired = any(
            error.get("extensions", {}).get("code") == "AUTH_TOKEN_EXPIRED" for error in errors
        )
        if token_is_expired:
            if self._refresh_auth_token():
                result = self._request_graphql(*args, **kwargs)
        return result


class ForestConnection:
    @_record_call
    def __init__(
        self,
        sync_endpoint: Optional[str] = None,
        compiler_endpoint: Optional[str] = None,
        forest_cloud_endpoint: Optional[str] = None,
    ):
        """
        Represents a connection to Forest containing methods to wrap all possible API endpoints.

        Users should not use methods from this class directly.

        :param sync_endpoint: The endpoint of the server for running QVM jobs
        :param compiler_endpoint: The endpoint of the server for running quilc compiler jobs
        :param forest_cloud_endpoint: The endpoint of the forest cloud server
        """
        pyquil_config = PyquilConfig()
        if sync_endpoint is None:
            sync_endpoint = pyquil_config.qvm_url
        if compiler_endpoint is None:
            compiler_endpoint = pyquil_config.quilc_url
        if forest_cloud_endpoint is None:
            forest_cloud_endpoint = pyquil_config.forest_url

        assert sync_endpoint is not None
        self.sync_endpoint = sync_endpoint
        self.compiler_endpoint = compiler_endpoint
        self.forest_cloud_endpoint = forest_cloud_endpoint
        self.session = get_session(config=pyquil_config)

    @_record_call
    def _run_and_measure(
        self, quil_program: Program, qubits: Sequence[int], trials: int, random_seed: int
    ) -> np.ndarray:
        """
        Run a Forest ``run_and_measure`` job.

        Users should use :py:func:`WavefunctionSimulator.run_and_measure` instead of calling
        this directly.
        """
        payload = run_and_measure_payload(quil_program, qubits, trials, random_seed)
        response = post_json(self.session, self.sync_endpoint + "/qvm", payload)
        return np.asarray(response.json())

    @_record_call
    def _wavefunction(self, quil_program: Program, random_seed: int) -> Wavefunction:
        """
        Run a Forest ``wavefunction`` job.

        Users should use :py:func:`WavefunctionSimulator.wavefunction` instead of calling
        this directly.
        """

        payload = wavefunction_payload(quil_program, random_seed)
        response = post_json(self.session, self.sync_endpoint + "/qvm", payload)
        return Wavefunction.from_bit_packed_string(response.content)

    @_record_call
    def _expectation(
        self, prep_prog: Program, operator_programs: Iterable[Program], random_seed: int
    ) -> np.ndarray:
        """
        Run a Forest ``expectation`` job.

        Users should use :py:func:`WavefunctionSimulator.expectation` instead of calling
        this directly.
        """
        if isinstance(operator_programs, Program):
            warnings.warn(
                "You have provided a Program rather than a list of Programs. The results "
                "from expectation will be line-wise expectation values of the "
                "operator_programs.",
                SyntaxWarning,
            )

        payload = expectation_payload(prep_prog, operator_programs, random_seed)
        response = post_json(self.session, self.sync_endpoint + "/qvm", payload)
        return np.asarray(response.json())

    @_record_call
    def _qvm_run(
        self,
        quil_program: Program,
        classical_addresses: Dict[str, Union[bool, Sequence[int]]],
        trials: int,
        measurement_noise: Optional[Tuple[float, float, float]],
        gate_noise: Optional[Tuple[float, float, float]],
        random_seed: Optional[int],
    ) -> np.ndarray:
        """
        Run a Forest ``run`` job on a QVM.

        Users should use :py:func:`QVM.run` instead of calling this directly.
        """
        payload = qvm_run_payload(
            quil_program, classical_addresses, trials, measurement_noise, gate_noise, random_seed
        )
        response = post_json(self.session, self.sync_endpoint + "/qvm", payload)

        ram = response.json()

        for k in ram.keys():
            ram[k] = np.array(ram[k])

        return ram

    @_record_call
    def _qvm_get_version_info(self) -> str:
        """
        Return version information for the QVM.

        :return: String of QVM version
        """
        response = post_json(self.session, self.sync_endpoint, {"type": "version"})
        split_version_string = response.text.split()
        try:
            qvm_version = split_version_string[0]
        except ValueError:
            raise TypeError(f"Malformed version string returned by the QVM: {response.text}")
        return qvm_version


class Engagement:
    """
    An Engagement stores all the information retrieved via an engagement request sent to
      the dispatch server.
    """

    def __init__(
        self,
        client_public_key: bytes,
        client_secret_key: bytes,
        server_public_key: bytes,
        expires_at: Union[int, float, str],
        qpu_endpoint: str,
        qpu_compiler_endpoint: str,
    ):
        self.client_public_key = client_public_key
        self.client_secret_key = client_secret_key
        self.server_public_key = server_public_key
        self.expires_at = float(expires_at) if expires_at else None
        self.qpu_endpoint = qpu_endpoint
        self.qpu_compiler_endpoint = qpu_compiler_endpoint
        logger.debug("New engagement created: \n%s", self)

    def is_valid(self) -> bool:
        """
        Return true if an engagement is valid for use, false if it is missing required
          fields

        An 'invalid' engagement is one which will not grant access to the QPU.
        """
        return all(
            [
                self.client_public_key is not None,
                self.client_secret_key is not None,
                self.server_public_key is not None,
                (self.expires_at is None or self.expires_at > time.time()),
                self.qpu_endpoint is not None,
            ]
        )

    def __str__(self) -> str:
        return f"""Client public key: {self.client_public_key}
Client secret key: masked ({len(self.client_secret_key)} B)
Server public key: {self.server_public_key}
Expiration time: {self.expires_at}
QPU Endpoint: {self.qpu_endpoint}
QPU Compiler Endpoint: {self.qpu_compiler_endpoint}"""  # type: ignore



##############################################################################
# Copyright 2016-2019 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

from typing import Sequence

import numpy as np

from pyquil.gates import MEASURE, RX, RZ
from pyquil.quil import Program


def parameterized_euler_rotations(
    qubits: Sequence[int],
    *,
    prefix: str,
    suffix_alpha: str = "alpha",
    suffix_beta: str = "beta",
    suffix_gamma: str = "gamma",
) -> Program:
    """
    Given a number of qubits (n), build a ``Program`` containing a ZXZXZ-decomposed gate on each
    qubit, where each ``RZ`` is parameterized by declared values with labels given by the "prefix"
    and "suffix" arguments. Put more plainly, the resulting Quil program on n qubits is::

        RZ(alpha_label[0]) 0
        RX(pi/2) 0
        RZ(beta_label[0]) 0
        RX(-pi/2) 0
        RZ(gamma_label[0]) 0
        ...
        RZ(alpha_label[n-1]) n-1
        RX(pi/2) n-1
        RZ(beta_label[0]) n-1
        RX(-pi/2) n-1
        RZ(gamma_label[n-1]) n-1

    :param qubits: The number of qubits (n).
    :param prefix: The prefix for the declared memory region labels. For example, if the prefix
        is "preparation" and the alpha, beta, and gamma suffixes are left as default, the labels
        would be "preparation_alpha", "preparation_beta", and "preparation_gamma".
    :param suffix_alpha: The suffix for the "alpha" memory region label, which corresponds to the
        first (rightmost) ``Z`` in the ZXZXZ decomposition. Defaults to "alpha".
    :param suffix_beta: The suffix for the "beta" memory region label, which corresponds to the
        second (middle) ``Z`` in the ZXZXZ decomposition. Defaults to "beta".
    :param suffix_gamma: The suffix for the "gamma" memory region label, which corresponds to the
        last (leftmost) ``Z`` in the ZXZXZ decomposition. Defaults to "gamma".
    :return: A ``Program`` containing a 3 parameterized ``RZ``s and 2 fixed ``RX``s per qubit.
    """
    alpha_label = f"{prefix}_{suffix_alpha}"
    beta_label = f"{prefix}_{suffix_beta}"
    gamma_label = f"{prefix}_{suffix_gamma}"

    p = Program()

    alpha = p.declare(alpha_label, "REAL", len(qubits))
    beta = p.declare(beta_label, "REAL", len(qubits))
    gamma = p.declare(gamma_label, "REAL", len(qubits))

    for idx, q in enumerate(qubits):
        p += RZ(alpha[idx], q)
        p += RX(np.pi / 2, q)
        p += RZ(beta[idx], q)
        p += RX(-np.pi / 2, q)
        p += RZ(gamma[idx], q)

    return p


def parameterized_single_qubit_state_preparation(
    qubits: Sequence[int], label: str = "preparation"
) -> Program:
    """
    Given a number of qubits, produce a program as in ``parameterized_euler_rotations`` where each
    memory region is prefixed by ``label``, where label defaults to "preparation".

    :param qubits: The number of qubits (n).
    :param label: The prefix to use when declaring memory in ``parameterized_euler_rotations``.
    :return: A parameterized ``Program`` that can be used to prepare a product state.
    """
    return parameterized_euler_rotations(qubits, prefix=label)


def parameterized_single_qubit_measurement_basis(
    qubits: Sequence[int], label: str = "measurement"
) -> Program:
    """
    Given a number of qubits, produce a program as in ``parameterized_euler_rotations`` where each
    memory region is prefixed by ``label``, where label defaults to "measurement".

    :param qubits: The number of qubits (n).
    :param label: The prefix to use when declaring memory in ``parameterized_euler_rotations``.
    :return: A parameterized ``Program`` that can be used to measure in a multi-qubit Pauli basis.
    """
    return parameterized_euler_rotations(qubits, prefix=label)


def parameterized_readout_symmetrization(
    qubits: Sequence[int], label: str = "symmetrization"
) -> Program:
    """
    Given a number of qubits (n), produce a ``Program`` with an ``RX`` instruction on qubits
    0 through n-1, parameterized by memory regions label[0] through label[n-1], where "label"
    defaults to "symmetrization".

    :param qubits: The number of qubits (n).
    :param label: The name of the declared memory region.
    :return: A ``Program`` with parameterized ``RX`` gates on n qubits.
    """
    p = Program()
    symmetrization = p.declare(f"{label}", "REAL", len(qubits))
    for idx, q in enumerate(qubits):
        p += RX(symmetrization[idx], q)
    return p


def measure_qubits(qubits: Sequence[int]) -> Program:
    """
    Given a number of qubits (n), produce a ``Program`` with a ``MEASURE`` instruction on qubits
    0 through n-1, with corresponding readout registers ro[0] through ro[n-1].

    :param qubits: The number of qubits (n).
    :return: A ``Program`` that measures n qubits.
    """
    p = Program()
    ro = p.declare("ro", "BIT", len(qubits))
    for idx, q in enumerate(qubits):
        p += MEASURE(q, ro[idx])
    return p



##############################################################################
# Copyright 2016-2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
import json
import warnings
from configparser import ConfigParser, NoSectionError, NoOptionError
from os import environ, path
from os.path import expanduser, abspath
from types import MappingProxyType
from typing import Callable, Dict, Iterable, Mapping, Optional, TYPE_CHECKING, cast

from pyquil.api._errors import UserMessageError
from pyquil.api._logger import logger

if TYPE_CHECKING:
    from pyquil.api._base_connection import Engagement

# `.qcs_config` is for content (mostly) related to QCS: the QCS front end stacks endpoint (`url`)
# for querying all QCS data: devices, reservations, etc. etc., and the `exec_on_engage` that the
# daemon will call during an engage event
QCS_CONFIG = "QCS_CONFIG"

# `.forest_config`, for content related to the Forest SDK, such as ip addresses for the various
# servers to which users submit quil & jobs (qvm, compiler, qpu, etc.)
FOREST_CONFIG = "FOREST_CONFIG"
CONFIG_PATHS = cast(
    Mapping[str, str],
    MappingProxyType({QCS_CONFIG: "~/.qcs_config", FOREST_CONFIG: "~/.forest_config"}),
)


class PyquilConfig(object):
    """
    The PyQuilConfig object holds the configuration necessary to communicate with Rigetti systems
        such as authentication tokens, endpoint URLs, and file paths for configuration files
        on disk.

    :attribute get_engagement: A callback to fetch a currently valid engagement from which to read
        configuration parameters (i.e., QPU_URL) as needed. This allows the engagement to be fetched
        and maintained elsewhere (i.e., by ForestSession or manually).
    """

    FOREST_URL = {
        "env": "FOREST_SERVER_URL",
        "file": QCS_CONFIG,
        "section": "Rigetti Forest",
        "name": "url",
        "default": "https://forest-server.qcs.rigetti.com",
    }

    API_KEY = {
        "env": "FOREST_API_KEY",
        "file": QCS_CONFIG,
        "section": "Rigetti Forest",
        "name": "key",
        "default": None,
    }

    USER_ID = {
        "env": "FOREST_USER_ID",
        "file": QCS_CONFIG,
        "section": "Rigetti Forest",
        "name": "user_id",
        "default": None,
    }

    DISPATCH_URL = {
        "env": "FOREST_DISPATCH_URL",
        "file": QCS_CONFIG,
        "section": "Rigetti Forest",
        "name": "dispatch_url",
        "default": "https://dispatch.services.qcs.rigetti.com/graphql",
    }

    ENGAGE_CMD = {
        "env": "QMI_ENGAGE_CMD",
        "file": QCS_CONFIG,
        "section": "QPU",
        "name": "exec_on_engage",
        "default": "",
    }

    QMI_AUTH_TOKEN_PATH = {
        "env": "QMI_AUTH_TOKEN_PATH",
        "file": QCS_CONFIG,
        "section": "Rigetti Forest",
        "name": "qmi_auth_token_path",
        "default": "~/.qcs/qmi_auth_token",
    }

    USER_AUTH_TOKEN_PATH = {
        "env": "USER_AUTH_TOKEN_PATH",
        "file": QCS_CONFIG,
        "section": "Rigetti Forest",
        "name": "user_auth_token_path",
        "default": "~/.qcs/user_auth_token",
    }

    QCS_URL = {
        "env": "QCS_URL",
        "file": FOREST_CONFIG,
        "section": "Rigetti Forest",
        "name": "qcs_url",
        "default": "https://qcs.rigetti.com",
    }

    QPU_URL = {
        "env": "QPU_URL",
        "file": FOREST_CONFIG,
        "section": "Rigetti Forest",
        "name": "qpu_endpoint_address",
        "engagement_key": "qpu_endpoint",
        "default": None,
    }

    QVM_URL = {
        "env": "QVM_URL",
        "file": FOREST_CONFIG,
        "section": "Rigetti Forest",
        "name": "qvm_address",
        "default": "http://127.0.0.1:5000",
    }

    QUILC_URL = {
        "env": "QUILC_URL",
        "file": FOREST_CONFIG,
        "section": "Rigetti Forest",
        "name": "quilc_address",
        "default": "tcp://127.0.0.1:5555",
    }

    QPU_COMPILER_URL = {
        "env": "QPU_COMPILER_URL",
        "file": FOREST_CONFIG,
        "section": "Rigetti Forest",
        "name": "qpu_compiler_address",
        "engagement_key": "qpu_compiler_endpoint",
        "default": None,
    }

    def __init__(self, config_paths: Mapping[str, str] = CONFIG_PATHS):
        """
        :param config_paths: The paths to the various configuration files read by pyQuil.
        """

        # The engagement callback can be added by config consumers after construction
        self.get_engagement: Optional[Callable[..., Optional[Engagement]]] = lambda: None

        # Whether engagement has been requested in order to provide any config values
        self._engagement_requested = False

        self.config_parsers: Dict[str, ConfigParser] = {}
        for env_name, default_path in config_paths.items():
            default_path = expanduser(default_path)
            path = environ.get(env_name, default_path)

            cp = ConfigParser()
            cp.read(abspath(path))
            self.config_parsers[env_name] = cp
        self._parse_auth_tokens()

    def _parse_auth_tokens(self) -> None:
        self.user_auth_token = _parse_auth_token(
            self.user_auth_token_path, ["access_token", "refresh_token", "scope"]
        )
        self.qmi_auth_token = _parse_auth_token(
            self.qmi_auth_token_path, ["access_token", "refresh_token"]
        )

    def _env_or_config_or_default(
        self,
        env: Optional[str] = None,
        file: Optional[str] = None,
        section: Optional[str] = None,
        name: Optional[str] = None,
        default: Optional[str] = None,
        engagement_key: Optional[str] = None,
    ) -> Optional[str]:
        """
        Get the value of the environment variable or config file value.
        The environment variable takes precedence.

        :param env: The environment variable name.
        :param name: The config file key.
        :param engagement_key: The attribute name by which this value can be read from
            an engagement. If None, then this value is not provided by engagement.
        :return: The value or None if not found
        """

        # If it's an envvar, that overrides all other options
        assert env is not None
        env_val = environ.get(env)
        if env_val is not None:
            return env_val

        # Otherwise, use the values in the config files from disk
        assert file is not None and section is not None and name is not None
        try:
            return self.config_parsers[file].get(section, name)
        except (NoSectionError, NoOptionError, KeyError):
            pass
        """
        If no local configuration is available, certain values are provided
            by the dispatch service.
        """
        assert self.get_engagement is not None
        try:
            if engagement_key is not None and self.get_engagement() is not None:
                self._engagement_requested = True
                return cast(str, getattr(self.get_engagement(), engagement_key))
        except AttributeError:
            pass

        return default

    @property
    def api_key(self) -> Optional[str]:
        warnings.warn(
            "PyQuilConfig#api_key is deprecated, unused by Rigetti services, and will "
            "be removed in a future release.",
            category=FutureWarning,
        )
        return self._env_or_config_or_default(**self.API_KEY)

    @property
    def dispatch_url(self) -> Optional[str]:
        return self._env_or_config_or_default(**self.DISPATCH_URL)

    @property
    def engage_cmd(self) -> Optional[str]:
        return self._env_or_config_or_default(**self.ENGAGE_CMD)

    @property
    def engagement(self) -> Optional["Engagement"]:
        """
        An Engagement should only be made available to consumers if it was used to retrieve a
        configuration value.
        """
        if not self._engagement_requested:
            return None
        assert self.get_engagement is not None
        return self.get_engagement()

    @property
    def forest_url(self) -> Optional[str]:
        return self._env_or_config_or_default(**self.FOREST_URL)

    @property
    def qmi_auth_token_path(self) -> str:
        value = self._env_or_config_or_default(**self.QMI_AUTH_TOKEN_PATH)
        assert value is not None
        return path.expanduser(value)

    @property
    def qcs_auth_headers(self) -> Dict[str, str]:
        if self.user_auth_token is not None:
            return {"Authorization": f'Bearer {self.user_auth_token["access_token"]}'}
        if self.qmi_auth_token is not None:
            return {"X-QMI-AUTH-TOKEN": self.qmi_auth_token["access_token"]}
        return {}

    @property
    def qcs_url(self) -> Optional[str]:
        return self._env_or_config_or_default(**self.QCS_URL)

    @property
    def qpu_compiler_url(self) -> Optional[str]:
        return self._env_or_config_or_default(**self.QPU_COMPILER_URL)

    @property
    def qpu_url(self) -> Optional[str]:
        return self._env_or_config_or_default(**self.QPU_URL)

    @property
    def quilc_url(self) -> Optional[str]:
        return self._env_or_config_or_default(**self.QUILC_URL)

    @property
    def qvm_url(self) -> Optional[str]:
        return self._env_or_config_or_default(**self.QVM_URL)

    def update_user_auth_token(self, user_auth_token: Dict[str, str]) -> None:
        self.user_auth_token = user_auth_token
        with open(self.user_auth_token_path, "w") as f:
            json.dump(user_auth_token, f)

    def update_qmi_auth_token(self, qmi_auth_token: Dict[str, str]) -> None:
        self.qmi_auth_token = qmi_auth_token
        with open(self.qmi_auth_token_path, "w") as f:
            json.dump(qmi_auth_token, f)

    @property
    def user_auth_token_path(self) -> str:
        value = self._env_or_config_or_default(**self.USER_AUTH_TOKEN_PATH)
        assert value is not None
        return path.expanduser(value)

    @property
    def user_id(self) -> Optional[str]:
        return self._env_or_config_or_default(**self.USER_ID)

    def assert_valid_auth_credential(self) -> None:
        """
        assert_valid_auth_credential will check to make sure the user has a valid
        auth credential configured. This assertion is made lazily - it is called
        only after the user has received a 401 or 403 from forest server. See
        _base_connection.py::ForestSession#_refresh_auth_token.
        """
        if self.user_auth_token is None and self.qmi_auth_token is None:
            raise UserMessageError(
                f"Your configuration does not have valid authentication credentials."
                f"Please visit {self.qcs_url}/auth/token to download credentials"
                f"and save to {self.user_auth_token_path}."
            )


def _parse_auth_token(path: str, required_keys: Iterable[str]) -> Optional[Dict[str, str]]:
    try:
        with open(abspath(expanduser(path)), "r") as f:
            token: Dict[str, str] = json.load(f)
            invalid_values = [k for k in required_keys if not isinstance(token.get(k), str)]
            if len(invalid_values) == 0:
                return token
            logger.warning(f"Failed to parse auth token at {path}.")
            logger.warning(f"Invalid {invalid_values}.")
    except json.decoder.JSONDecodeError:
        logger.warning(f"Failed to parse auth token at {path}. Invalid JSON.")
    except FileNotFoundError:
        logger.debug("Auth token at %s not found.", path)
    return None
